@@ -96,6 +96,7 @@
 
 /* General macros */
 #define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))
+#define sizeof_field(type, member) sizeof((((type *)0)->member))
 
 /** Bits per byte.
  * Use this instead of a magic constant to illustrate why something
@@ -134,6 +135,8 @@ enum kdump_arch {
 	ARCH_MIPS,
 	ARCH_PPC,
 	ARCH_PPC64,
+	ARCH_RISCV32,
+	ARCH_RISCV64,
 	ARCH_S390,
 	ARCH_S390X,
 	ARCH_X86_64,
@@ -709,6 +712,9 @@ struct _kdump_ctx {
 	/** Per-context data. */
 	void *data[PER_CTX_SLOTS];
 
+	/** Temporary buffer for file names in error messages. */
+	char err_filename[sizeof("File #") + 20];
+
 	/** Error message buffer.
 	 * This must be the last member. */
 	kdump_errmsg_t err;
@@ -739,7 +745,9 @@ INTERNAL_DECL(kdump_status, linux_iomem_kcode,
 /* Architectures */
 
 INTERNAL_DECL(extern const struct arch_ops, aarch64_ops, );
+INTERNAL_DECL(extern const struct arch_ops, arm_ops, );
 INTERNAL_DECL(extern const struct arch_ops, ia32_ops, );
+INTERNAL_DECL(extern const struct arch_ops, riscv64_ops, );
 INTERNAL_DECL(extern const struct arch_ops, s390x_ops, );
 INTERNAL_DECL(extern const struct arch_ops, x86_64_ops, );
 INTERNAL_DECL(extern const struct arch_ops, ppc64_ops, );
@@ -770,6 +778,9 @@ INTERNAL_DECL(kdump_status, set_error,
 	      (kdump_ctx_t *ctx, kdump_status ret, const char *msgfmt, ...))
 	__attribute__ ((format (printf, 3, 4)));
 
+INTERNAL_DECL(const char *, err_filename,
+	      (kdump_ctx_t *ctx, unsigned fidx));
+
 INTERNAL_DECL(kdump_status, addrxlat2kdump,
 	      (kdump_ctx_t *ctx, addrxlat_status status));
 INTERNAL_DECL(addrxlat_status, kdump2addrxlat,
@@ -797,6 +808,16 @@ INTERNAL_DECL(kdump_status, get_symbol_val,
 INTERNAL_DECL(kdump_status, set_file_description,
 	      (kdump_ctx_t *ctx, const char *name));
 
+INTERNAL_DECL(kdump_status, set_blob_attr,
+	      (kdump_ctx_t *ctx, enum global_keyidx attr,
+	       void *data, size_t size, const char *desc));
+INTERNAL_DECL(kdump_status, read_blob_attr,
+	      (kdump_ctx_t *ctx, unsigned fidx, off_t off, size_t size,
+	       enum global_keyidx attr, const char *desc));
+
+INTERNAL_DECL(kdump_status, read_current_vmcoreinfo,
+	      (kdump_ctx_t *ctx));
+
 /**  Definition of a derived attribute.
  *
  * This structure is used to translate raw binary data to an attribute
@@ -813,6 +834,15 @@ struct derived_attr_def {
 	unsigned short length;
 };
 
+#define DERIVED_NUMBER(k, d, t, f) {	\
+	{ .key = k,			\
+	  { .depth = d },		\
+	  .type = KDUMP_NUMBER		\
+	},				\
+	.offset = offsetof(t, f),	\
+	.length = sizeof_field(t, f)	\
+}
+
 /** Get pointer to the definition of a derived attribute.
  *
  * This macro only works for derived attributes which are embedded inside
@@ -836,8 +866,7 @@ INTERNAL_DECL(kdump_status, create_xen_cpu_regs,
 	       struct derived_attr_def *def, unsigned ndef));
 
 /* unaligned access */
-struct unaligned_long
-{
+struct unaligned_long {
 	unsigned long val;
 } __attribute__((packed));
 
@@ -847,6 +876,15 @@ get_unaligned_long(const unsigned char *p)
 	return ((struct unaligned_long*)p)->val;
 }
 
+struct unaligned_uint64_t {
+	uint64_t val;
+} __attribute__((packed));
+
+static inline uint64_t
+get_unaligned_uint64_t(const unsigned char *p)
+{
+	return ((struct unaligned_uint64_t*)p)->val;
+}
 
 /* hashing */
 INTERNAL_DECL(unsigned long, mem_hash, (const char *s, size_t len));
