@@ -44,6 +44,8 @@
 # include <zlib.h>
 #endif
 
+#define FN_VMCOREINFO	"/sys/kernel/vmcoreinfo"
+
 /** Set an error message, returning @c kdump_status.
  * @arg err     Error message object.
  * @arg status  Status code.
@@ -75,6 +77,8 @@ kdump_err(kdump_ctx_t *ctx, kdump_status status, const char *msgfmt, ...)
 	if (status != KDUMP_OK) {
 		va_list ap;
 
+		if (status == KDUMP_ERR_SYSTEM && !err_str(&ctx->err))
+			err_add(&ctx->err, "%s", strerror(errno));
 		va_start(ap, msgfmt);
 		err_vadd(&ctx->err, msgfmt, ap);
 		va_end(ap);
@@ -201,6 +205,7 @@ arch_ptr_size(enum kdump_arch arch)
 	case ARCH_ALPHA:
 	case ARCH_IA64:
 	case ARCH_PPC64:
+	case ARCH_RISCV64:
 	case ARCH_S390X:
 	case ARCH_X86_64:
 		return 8;	/* 64 bits */
@@ -209,6 +214,7 @@ arch_ptr_size(enum kdump_arch arch)
 	case ARCH_IA32:
 	case ARCH_MIPS:
 	case ARCH_PPC:
+	case ARCH_RISCV32:
 	case ARCH_S390:
 	default:
 		return 4;	/* 32 bits */
@@ -234,6 +240,10 @@ machine_arch_name(const char *machine)
 	else if (!strcmp(machine, "ppc64") ||
 		 !strcmp(machine, "ppc64le"))
 		return KDUMP_ARCH_PPC64;
+	else if (!strcmp(machine, "riscv32"))
+		return KDUMP_ARCH_RISCV32;
+	else if (!strcmp(machine, "riscv64"))
+		return KDUMP_ARCH_RISCV64;
 	else if (!strcmp(machine, "s390"))
 		return KDUMP_ARCH_S390;
 	else if (!strcmp(machine, "s390x"))
@@ -264,6 +274,8 @@ default_page_shift(enum kdump_arch arch)
 		[ARCH_MIPS] = 12,
 		[ARCH_PPC] = 0,
 		[ARCH_PPC64] = 0,
+		[ARCH_RISCV32] = 12,
+		[ARCH_RISCV64] = 12,
 		[ARCH_S390] = 12,
 		[ARCH_S390X] = 12,
 		[ARCH_IA32] = 12,
@@ -280,17 +292,19 @@ arch_ops(enum kdump_arch arch)
 {
 	switch (arch) {
 	case ARCH_ALPHA:
-	case ARCH_ARM:
 	case ARCH_IA64:
 	case ARCH_MIPS:
 	case ARCH_PPC:
+	case ARCH_RISCV32:
 	case ARCH_S390:
 		/* TODO */
 		break;
 
 	case ARCH_AARCH64:	return &aarch64_ops;
+	case ARCH_ARM:		return &arm_ops;
 	case ARCH_IA32:		return &ia32_ops;
 	case ARCH_PPC64:	return &ppc64_ops;
+	case ARCH_RISCV64:	return &riscv64_ops;
 	case ARCH_S390X:	return &s390x_ops;
 	case ARCH_X86_64:	return &x86_64_ops;
 
@@ -315,6 +329,8 @@ static const char canon_arch_names[][MAX_ARCH_NAME_LEN] =
 	DEF_ARCH(MIPS),
 	DEF_ARCH(PPC),
 	DEF_ARCH(PPC64),
+	DEF_ARCH(RISCV32),
+	DEF_ARCH(RISCV64),
 	DEF_ARCH(S390),
 	DEF_ARCH(S390X),
 	DEF_ARCH(X86_64),
@@ -1288,3 +1304,131 @@ set_file_description(kdump_ctx_t *ctx, const char *name)
 	return set_attr_static_string(ctx, gattr(ctx, GKI_file_description),
 				      ATTR_DEFAULT, name);
 }
+
+/** Set a blob attribute.
+ * @param ctx   Dump file object.
+ * @param attr  Global key idx.
+ * @param data  Raw data.
+ * @param size  Size of raw data.
+ * @param desc  Human-readable name of the blob (for error messages).
+ */
+kdump_status
+set_blob_attr(kdump_ctx_t *ctx, enum global_keyidx attr,
+	       void *data, size_t size, const char *desc)
+{
+	kdump_attr_value_t val;
+	kdump_status res;
+
+	val.blob = internal_blob_new_dup(data, size);
+	if (!val.blob)
+		return set_error(ctx, KDUMP_ERR_SYSTEM,
+				 "Cannot allocate %s", desc);
+	res = set_attr(ctx, gattr(ctx, attr), ATTR_DEFAULT, &val);
+	if (res != KDUMP_OK)
+		return set_error(ctx, res, "Cannot set %s", desc);
+
+	return KDUMP_OK;
+}
+
+/** Read a blob attribute from file.
+ * @param ctx   Dump file object.
+ * @param fidx  File index.
+ * @param off   File offset.
+ * @param size  Size of the blob.
+ * @param attr  Global attribute key index.
+ * @param desc  Human-readable name of the blob.
+ */
+kdump_status
+read_blob_attr(kdump_ctx_t *ctx, unsigned fidx, off_t off, size_t size,
+	       enum global_keyidx attr, const char *desc)
+{
+	struct fcache_chunk fch;
+	kdump_status ret;
+
+	ret = fcache_get_chunk(ctx->shared->fcache, &fch, size, fidx, off);
+	if (ret != KDUMP_OK)
+		return set_error(ctx, ret,
+				 "Cannot read %s (%zu bytes at %llu in %s)",
+				 desc, size, (unsigned long long) off,
+				 err_filename(ctx, fidx));
+
+	ret = set_blob_attr(ctx, attr, fch.data, size, desc);
+
+	fcache_put_chunk(&fch);
+	return ret;
+}
+
+/** Read running kernel's VMCOREINFO from
+ * @param ctx   Dump file object.
+ * @returns     Error status.
+ *
+ * Get VMCOREINFO location from /sys/kernel/vmcoreinfo and parse it
+ * using the currently opened dump file (which should be a live dump).
+ */
+kdump_status
+read_current_vmcoreinfo(kdump_ctx_t *ctx)
+{
+	FILE *f;
+	unsigned long long addr;
+	size_t length;
+	void *info;
+	kdump_status ret;
+
+	f = fopen(FN_VMCOREINFO, "r");
+	if (!f)
+		return errno == ENOENT
+			? KDUMP_OK
+			: set_error(ctx, KDUMP_ERR_SYSTEM,
+				    "Cannot open %s", FN_VMCOREINFO);
+
+	if (fscanf(f, "%llx %zx", &addr, &length) == 2)
+		ret = KDUMP_OK;
+	else if (ferror(f))
+		ret = set_error(ctx, KDUMP_ERR_SYSTEM,
+				"Error reading %s", FN_VMCOREINFO);
+	else
+		ret = set_error(ctx, KDUMP_ERR_CORRUPT,
+				"Error parsing %s: Wrong file format",
+				FN_VMCOREINFO);
+	fclose(f);
+	if (ret != KDUMP_OK)
+		return ret;
+
+	info = ctx_malloc(length, ctx, "VMCOREINFO buffer");
+	if (!info)
+		return KDUMP_ERR_SYSTEM;
+
+	ret = read_locked(ctx, KDUMP_MACHPHYSADDR, addr, info, &length);
+	if (ret == KDUMP_OK)
+		ret = process_notes(ctx, info, length);
+
+	free(info);
+	return ret;
+}
+
+/** Get file name for error messages.
+ * @param ctx   Dump file object.
+ * @param fidx  File index.
+ * @returns     Error status.
+ *
+ * The file name buffer is pre-allocated in @p ctx, so the returned
+ * string stays valid as long as @p ctx stays valid, or until the
+ * next call to this function.
+ */
+const char *
+err_filename(kdump_ctx_t *ctx, unsigned fidx)
+{
+	struct attr_data *attr;
+	const char *numstr;
+
+	sprintf(ctx->err_filename, "file #%u", fidx);
+	numstr = ctx->err_filename + sizeof("file #") - 1;
+	attr = lookup_dir_attr(ctx->dict, gattr(ctx, GKI_dir_file_set),
+			       numstr, strlen(numstr));
+	if (attr) {
+		attr = lookup_dir_attr(ctx->dict, attr, "name", 4);
+		if (attr && attr_isset(attr))
+			return attr_value(attr)->string;
+	}
+	return ctx->err_filename;
+}
