@@ -488,7 +488,8 @@ vtop_init(kdump_ctx_t *ctx)
 	if (!attr_isset(gattr(ctx, GKI_pteval_size)))
 		set_pteval_size(ctx);
 
-	if (ctx->shared->arch_ops->post_addrxlat &&
+	if (ctx->shared->arch_ops &&
+	    ctx->shared->arch_ops->post_addrxlat &&
 	    (status = ctx->shared->arch_ops->post_addrxlat(ctx)) != KDUMP_OK)
 		return set_error(ctx, status,
 				 "Arch late init failed");
@@ -749,7 +750,7 @@ sym_offsetof_locked(const addrxlat_cb_t *cb, const char *obj, const char *elem,
 	if (!attr)
 		return ADDRXLAT_ERR_NODATA;
 
-	attr = sub_attr_xlat(ctx, attr, obj, "Field");
+	attr = sub_attr_xlat(ctx, attr, elem, "Field");
 	if (!attr)
 		return ADDRXLAT_ERR_NODATA;
 
